/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "ORBTestClient.h"
#import <ADORB/ORB.h>
#import "TestProto.h"
#import "NamingContextProto.h"

#define MAP_OCTET_TO_NSDATA 0

@interface NSObject(Server)

- (double) addDouble:(double)a b:(double)b;
- (long long) addLongLong:(long long)a b:(long long)b;
- (void) addToLongLong:(long long*)a b:(long long)b;
- (NSArray*) queryEvents: the_area
    minDepth: min_depth
    maxDepth: max_depth
    timeRange: time_range
    searchTypes:(NSArray*)search_types
    minMagnitude:(float) min_magnitude
    maxMagnitude:(float) max_magnitude
    catalogs:(NSArray*) catalogs
    contributors:(NSArray*) contributors
    seqMax:(long) seq_max
    iter:(id*)iter;
@end 

@implementation ORBTestClient

- init
{
#if MAP_OCTET_TO_NSDATA
	[ORB setMapsOctetToNSData:YES];
#endif
	NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
	// Get parameter whether to use NamingContextExt
	BOOL useNSExt = [defaults boolForKey:@"useNSExt"];
	NSString* nsURL = [defaults stringForKey:@"ORBNamingURL"];
	if(!nsURL){
		// Get value for orbd's ORBInitialPort
		NSString* nsPort = [defaults stringForKey:@"port"];
		if(!nsPort)
			nsPort = @"15600"; // The default orbd's ORBInitialPort is different
		// Form corbaloc URL
		nsURL = [NSString stringWithFormat:@"corbaloc:iiop:1.2@localhost:%@", nsPort];
	}
	// This is the path for our object in the naming service
	NSString* path = @"ORBTest";
    self = [super init];
	ORB* orb = [ORB defaultORB]; // force to parse IDLs
    @try{
		[CORBAObject setDefaultConnectTimeout:10.0];
		// connect to naming server (using port 15600, default port is different)
		CORBAObject* ns = [CORBAObject objectWithURL:nsURL];
		// corbaloc provides no information of what interface 'ns' implements
		// so we need to set interface explicitly
		if(useNSExt){
			// Get interface object for NamingContextExt and set it
			IDLInterface* nsInterface = [orb interfaceNamed:@"CosNaming/NamingContextExt"];
			[ns setInterface:nsInterface];
			server = [[ns resolveStr:path]retain];
		}else{
			// Get interface object for NamingContext and set it
			IDLInterface* nsInterface = [orb interfaceNamed:@"CosNaming/NamingContext"];
			[ns setInterface:nsInterface];
			server = [[ns resolve:[IDLNameComponent nameComponentsForPath:path]]retain];
		}
    }@catch(NSException *localException){
		NSLog(@"%@: %@\n", [localException name], [localException reason]);
    }
    return self;
}
- (void) dealloc
{
    [server release];
    [super dealloc];
}

- (void) sendLongDataWith:(NSString*)str
{
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
	@try{
		NSMutableData* data = [NSMutableData data];
		const char* bytes = [str UTF8String];
		int k;
		for(k=0;k<100000;k++)
			[data appendBytes:bytes length:[str length]];
		NSData* data1;
		NSLog(@"- sending %d bytes", [data length]);
		NSData* newData = [server echoSeqOctet:data result:&data1];
		if([newData isEqualToData:data])
			NSLog(@"- LongData OK");
		else
			NSLog(@"***** LongData failed *****");
	}@catch(NSException *localException){
		NSLog(@"***** LongData failed - %@, %@", [localException name], [localException reason]);
	}
	[pool release];
}

- (void) _run
{
    NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
    @try{

		short a = random()%16003, b = random()%16003, c, d;
		c = [server addShort:a b:b sum:&d];
		NSAssert(c == d && c == a + b, @"add_short");

		long k = random(), l = random(), n;
		long m = [server addLong:k b:l sum:&n];
		NSAssert(m == n && n == k + l, @"add_long");

		long long kk = 7, ll = 8, nn;
		long long mm = [server addLongLong:kk b:ll sum:&nn];
		NSAssert(mm == nn && nn == kk + ll, @"add_long_long");

		mm = kk = random();
		ll = random();
		[server addToLongLong:&kk b:ll];
		NSAssert(kk == ll + mm, @"add_to_long_long");
		
		double x = 5, y = 7, w, z;
		z = [server addDouble:x b:y sum:&w];
		NSAssert(z == x + y && w == z, @"add_double");

		float fx = 5, fy = 7, fz, fw;
		fz = [server addFloat:fx b:fy sum:&fw];
		NSAssert(fz == fx + fy && fw == fz, @"add_float");

		c = a;
		[server addToShort:&c b:b];
		NSAssert(c == a + b, @"add_to_short");

		z = x;
		[server addToDouble:&z b:y];
		NSAssert(z == x + y, @"add_to_double");

		fz = fx;
		[server addToFloat:&fz b:fy];
		NSAssert(z == x + y, @"addToDouble");
		
		x = 9;
		[server sqrt:x sqrt:&z];
		NSAssert(z == sqrt(x), @"sqrt");
		
		a = 1;
		c = [server echoShort:a result:&b];
		NSAssert(b == a && c == a, @"echo_short");
		
		char c0 = 'A',c1,c2;
		c2 = [server echoChar:c0 returnValue:&c1]; // selector from #pragma
		NSAssert( c0 == c1 && c1 == c2, @"echo_char");

#if MAP_OCTET_TO_NSDATA
		unsigned char u0 = '0xc7';
		NSData *data0 = [NSData dataWithBytes:&u0 length:1];
		NSData *data1, *data2;
		data2 = [server echoOctet:data0 returnValue:&data1];
		unsigned char *u1 = [data1 bytes], *u2 = [data2 bytes];
		NSAssert(u0 == u1[0] && u0 == u2[0], @"echo_octet");
		u1 = (u0>>4) & 0x0f | (u0<<4) & 0xf0;
		[server dupOctet:&data0];
		u2 = [data0 bytes];
		NSAssert(u1 == u2[0], @"dup_octet");
#else
		unsigned char u0 = '0xc7', u1, u2;
		u2 = [server echoOctet:u0 returnValue:&u1]; // selector from #pragma
		NSAssert(u0 == u1 && u1 == u2, @"echo_octet");
		
		u1 = (u0>>4) & 0x0f | (u0<<4) & 0xf0;
		[server dupOctet:&u0];
		NSAssert(u1 == u0, @"dup_octet");
#endif	
		NSMutableArray* array = [server getAllTests];
		NSAssert(server == [array objectAtIndex:0], @"getAllTests");
		
		NSString* s0 = @"abc", *s1, *s2;
		s2 = [server echoString:s0 result:&s1];
		NSAssert([s0 isEqualToString:s1] && [s1 isEqualToString:s2], @"echo_string");
		
		NSArray* context = [NSArray arrayWithObjects:
			[NSDictionary dictionaryWithObjectsAndKeys:
				@"ctx-value", @"value", @"ctx", @"name",nil],
			[NSDictionary dictionaryWithObjectsAndKeys:
				@"ctx2-value", @"value", @"ctx2", @"name", nil],
				nil];
		NSString* answer = [server whatIs:@"request" context:context];
		NSLog(@"return = %@\n", answer);
		
		fx = 15;
		[server setValue:fx];
		
		NSDate* date = [NSDate date];
		m = 0;
		for(k=0;k<100;k++)
			[server addToLong:&m value:k]; // selector from #pragma
		NSLog(@"sum(1..100) = %d, # of calls %d, duration %f", m, k, [[NSDate date]timeIntervalSinceDate:date]);
		
		[self sendLongDataWith:@"0123456789"];
		[NSThread detachNewThreadSelector:@selector(sendLongDataWith:) toTarget:self withObject:@"abcdefghij"];
		[NSThread detachNewThreadSelector:@selector(sendLongDataWith:) toTarget:self withObject:@"ABCDEFGHIJ"];
		
		array = [NSMutableArray array];
		for(k=1; k<100000;k++)
			[array addObject:[NSNumber numberWithLong:k]];
		mm = [server sum:array];
		NSLog(@"Sum(1 .. 100000) = %qi",mm);

		NSLog(@"*** ORBTest passed ***");
    }@catch(NSException *localException){
		NSLog(@"Exception %@: %@\n", [localException name], [localException reason]);
    }
    [pool release];
}

- (void) run
{
	NSLog(@"%@: starting...\n", [self className]);
    [self _run];
}


@end
