/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLDefinition_Private.h"
#import "IDLTypedef.h"
#import "IDLDefinitionScope_Private.h"
#import "IDLEncapsulation.h"
#import "IDLLink.h"
#import "IDLReference.h"

@implementation IDLTypedef

- initWithType:(id <IDLDefinition,NSObject>)type name:(NSString*)name
{
    self = [super init];
    _type = [type retain];
    _name = [name retain];
    _prefix = [[IDLDefinitionScope currentPrefix]retain];
    return self;
}

- _initWithCoder:(id<CORBADecoder>)coder
{
    self = [super init];
    @try{
		[coder decodeString:&_typeId];
		[_typeId retain];
		[coder decodeString:&_name];
		[_name retain];
		_type = [[IDLDefinition definitionFromCoder:coder]retain];
    }@catch(NSException *localException){
		[self release];
		[localException raise];
    }
    return self;
}

- (void) dealloc
{
    [_type release];
    [_name release];
    [_typeId release];
    [_prefix release];
    [super dealloc];
}

+ (NSZone*) zone
{
    return [IDLDefinition zone];
}

+ typedefWithType:(NSObject<IDLDefinition>*)type name:(NSString*)name
{
    return [[[self allocWithZone:[self zone]]initWithType:type name:name]autorelease];
}

+ _definitionFromCoder:(id<CORBADecoder>)coder
{
    IDLTypedef* def = [[self allocWithZone:[self zone]]autorelease];
    [coder registerDefinition:(IDLDefinition*)def forLocation:-2*sizeof(long)];
    return [def _initWithCoder:coder];
}

- (NSString*) name
{
    return _name;
}

- (void) forwardInvocation:(NSInvocation*)inv
{
    if([_type isKindOfClass:[IDLTypedef class]] ||
            [_type isKindOfClass:[IDLLink class]] ||
            [_type isKindOfClass:[IDLReference class]])
        [(IDLTypedef*)_type forwardInvocation:inv];
    else
    if([_type respondsToSelector:[inv selector]])
        [inv invokeWithTarget:_type];
    else
        [self doesNotRecognizeSelector:[inv selector]];
}

- (IDLDefinition*)referenceType
{
    return [_type referenceType];
}

- (BOOL) isDerivedFromIDLType:(IDLDefinition*)type
{
	return [[self referenceType]isDerivedFromIDLType:type];
}

- (NSMethodSignature*) methodSignatureForSelector:(SEL)sel
{
    return [(NSObject*)_type methodSignatureForSelector:sel];
}

- (NSString*) prefix
{
    return _prefix;
}

- (void) setNamePath:(NSString*)path
{
    NSString* prefix = [self prefix];
    path = path?[NSString stringWithFormat:@"%@/%@",path,[self name]] : [self name];
    [_typeId release];
    if([prefix length] > 0)
        _typeId = [[NSString stringWithFormat:@"IDL:%@/%@:1.0", prefix, path]retain];
    else
        _typeId = [[NSString stringWithFormat:@"IDL:%@:1.0", path]retain];
}

- (long) kind
{
    return tk_alias;
}

- (NSString*) typeId
{
    return _typeId;
}

- (void) _encodeWithCORBACoder:(id <CORBACoder>)coder
{
    [coder encodeString:[self typeId]];
    [coder encodeString:[self name]];
    [(IDLDefinition*)_type encodeWithCORBACoder:coder];
}

- (void) encodeWithCORBACoder:(id <CORBACoder>)coder
{
    long location = [coder locationForDefinition:(IDLDefinition*)self];
    if(location != 0 && [IDLDefinition allowsIndirection]) // already encoded -> encode indirection
        [coder encodeIndirection:location];
    else{
        IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
        [coder encodeLong:[self kind]];
        [stream setParentStream:coder location:[coder location]];
        [stream registerLocation:-2*sizeof(long) forDefinition:(IDLDefinition*)self];
        [self _encodeWithCORBACoder:stream];
        [coder encodeData:[stream data]];
    }
}

- (NSString*) version
{
    return [_type version];
}

- valueForKey:(NSString*)key
{
	return [_type valueForKey:key];
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	if([coder allowsKeyedCoding]){
		[coder encodeObject:_typeId forKey:@"typeId"];
		[coder encodeObject:_prefix forKey:@"prefix"];
		[coder encodeObject:_name forKey:@"name"];
		[coder encodeObject:_type forKey:@"type"];
	}else{
		[coder encodeObject:_typeId];
		[coder encodeObject:_prefix];
		[coder encodeObject:_name];
		[coder encodeObject:_type];
	}
}

- initWithCoder:(NSCoder*)coder
{
	self = [super init];
	if([coder allowsKeyedCoding]){
		_typeId = [[coder decodeObjectForKey:@"typeId"]retain];
		_prefix = [[coder decodeObjectForKey:@"prefix"]retain];
		_name = [[coder decodeObjectForKey:@"name"]retain];
		_type = [[coder decodeObjectForKey:@"type"]retain];
	}else{
		_typeId = [[coder decodeObject]retain];
		_prefix = [[coder decodeObject]retain];
		_name = [[coder decodeObject]retain];
		_type = [[coder decodeObject]retain];
	}
	return self;
}
- (id)awakeAfterUsingCoder:(NSCoder *)aDecoder
{
	id obj = [(NSObject*)[[IDLDefinitionScope currentScope]definitionForTypeId:_typeId]retain];
	if(obj && obj != self){
		[self release];
		return obj;
	}else
		return self;
}

@end
