/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "CORBAObject.h"
#import "CORBAObject_Private.h"
#import "VXInvocation.h"
#import "VXConnection.h"
#import "IDLNameComponent.h"
#import "ORB.h"
#import "ORB_Private.h"
#import <IDL.h>
#import "IDLGIOPVersion.h"
#import "IDLProfileBody.h"
#import "IDLIOPProfile.h"
#import "IDLTaggedComponent.h"
#import "IDLServiceContext.h"

@interface _CORBAObjectRef : NSObject
{
@public
	NSData* objectID;
	NSString* address;
	unsigned port;
	CORBAObject* object;
}
+ objectWithObjectID:(NSData*)objId address:(NSString*)addr port:(unsigned)port;
@end

@implementation CORBAObject

static NSMutableSet *_objects = nil;
static NSRecursiveLock* _hlock = nil;
NSString* ORBInvalidateObject = @"CORBAObjectInvalidatedNotification";
NSString* VXInvalidateObject = @"CORBAObjectInvalidatedNotification"; // deprecated
NSString* CORBAObjectDidConnect = @"CORBAObjectDidConnectNotification";
NSString* VXObjectDidConnect = @"CORBAObjectDidConnectNotification"; // deprecated
NSString* CORBAObjectDidNotConnect = @"CORBAObjectDidNotConnectNotification";
NSString* VXObjectDidNotConnect = @"CORBAObjectDidNotConnectNotification"; // deprecated
BOOL VXVerifyInterfaceVersion = YES;
NSString* InvalidFormatException = @"InvalidFormatException";
NSString* IDLDefinitionException = @"IDLDefinitionException";
NSString* CORBAAlreadyBoundException = @"CORBAAlreadyBoundException";
NSString* CORBANotActivatableException = @"CORBANotActivatableException";

NSString* CORBAErrorDomain = @"CORBAErrorDomain";

static NSMutableSet* _addressSet = nil;
static NSString* _address = nil;

static CFUUIDRef _uuidRef = NULL;
static NSData* _uuid = nil;
static double _defaultReplyTimeout = 300.0;
static double _defaultConnectTimeout = 75.0;

+ (void) initialize
{
    if([CORBAObject class] == self){
		_uuidRef = CFUUIDCreate(NULL);
		CFUUIDBytes bytes = CFUUIDGetUUIDBytes(_uuidRef);
		_uuid = [[NSData alloc]initWithBytesNoCopy:&bytes length:16];
		_objects = [[NSMutableSet alloc]init];
        _hlock = [[NSRecursiveLock alloc]init];
#if !(TARGET_IPHONE_SIMULATOR || TARGET_OS_IPHONE)
		[self setupIPMonitor];
#endif
    }
}

+ (double) defaultReplyTimeout
{
	return _defaultReplyTimeout;
}
+ (double) defaultConnectTimeout
{
	return _defaultConnectTimeout;
}
+ (void) setDefaultReplyTimeout:(double)timeout
{
	_defaultReplyTimeout = timeout;
}
+ (void) setDefaultConnectTimeout:(double)timeout
{
	_defaultConnectTimeout = timeout;
}

+ defaultIIOPVersion
{
    return [IDLGIOPVersion defaultIIOPVersion];
}

+ (void) setDefaultIIOPVersion:(unsigned)major:(unsigned)minor
{
    [IDLGIOPVersion setDefaultIIOPVersion:major:minor];
}

+ (void) setMaxMinorVersion:(unsigned)minor
{
    [IDLGIOPVersion setMaxMinorVersion:minor];
}

+ (NSString*) localHostAddress
{
	return _address;
}

+ (NSSet*) localHostAddressSet
{
	return _addressSet? [NSSet setWithSet:_addressSet] : [NSSet set];
}

- (IDLInterface*)interface
{
    return _interface;
}

- implementation
{
    return _implementation;
}

- (NSMethodSignature*)methodSignatureForSelector:(SEL)selector
{
    if([super respondsToSelector:selector])
        return [super methodSignatureForSelector:selector];
    else
    if(![self isLocal]){
        IDLFunction* fo = [[self interface]functionWithSelectorNamed:NSStringFromSelector(selector)];
        return [fo methodSignature];
    }else
        return [[self implementation]methodSignatureForSelector:selector];
}

- (void) forwardInvocation:(NSInvocation*)inv
{
    if(![self isLocal]){
        SEL selector = [inv selector];
        IDLFunction* fo = [[self interface]functionWithSelectorNamed:NSStringFromSelector(selector)];
        NSMutableDictionary* dict = [fo getArgValues:nil fromInvocation:inv];
        VXInvocation* cinv = [self createInvocationFor:fo withArguments:dict];
        [cinv invoke];
        [fo setValues:dict forInvocation:inv];
    }else
        [inv invokeWithTarget:[self implementation]];
}

- (void) _setInterface:(IDLInterface*)ifObj
{
    [_interface release];
    _interface = [ifObj retain];
//    _typeId = [[_interface typeId]retain];
}

- (void) setInterface:(IDLInterface*)newInterface
{
    if(_interface != newInterface){
		[_interface release];
        _interface = [newInterface retain];
		[_typeId release];
		_typeId = [[newInterface typeId]retain];
    }
}

- initWithInterface:(IDLInterface*)ifObj
{
    [super init];
    _interface = [ifObj retain];
    _typeId = [[_interface typeId]retain];
    _replyTimeout = _defaultReplyTimeout;
	_connectTimeout = _defaultConnectTimeout;
    return self;
}
- init
{
    self = [self initWithInterface:nil];
    return self;
}

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter]removeObserver:self];
	if(_delegate)
		[[NSNotificationCenter defaultCenter]removeObserver:_delegate name:nil object:self];
    if(_flags.unique){
        IDLProfileBody* data = [[self profileWithTag:TAG_INTERNET_IOP]profileBody];
#if USE_DEFS
		struct {@defs(_CORBAObjectRef)} ref;
		ref.objectID = [data objectID];
		ref.address = [data hostName];
		ref.port = [data portNumber];
		ref.isa = [_CORBAObjectRef class];
        [_hlock lock];
		[_objects removeObject:(id)&ref];
        [_hlock unlock];
#else
		_CORBAObjectRef *ref = [_CORBAObjectRef objectWithObjectID:[data objectID] address:[data hostName] port:[data portNumber]];
		[_hlock lock];
		[_objects removeObject:ref];
		[_hlock unlock];
#endif
    }
    [_typeId release];
    [_profiles release];
    [_interface release];
    [_connection release];
    [_implementation release];
	[_codeSetsContext release];
	[_error release];
    [super dealloc];
}

- (void) setIsUnique
{
    _flags.unique = YES;
}

+ (CORBAObject*) _lookupObject:(CORBAObject*)obj
{
    if(obj && [obj isKindOfClass:[CORBAObject class]]){
        IDLProfileBody* data = [[obj profileWithTag:TAG_INTERNET_IOP]profileBody];
		NSString* address = [data hostName];
#if USE_DEFS
		struct {@defs(_CORBAObjectRef)} ref;
		ref.objectID = [data objectID];
		ref.address = address;
		ref.port = [data portNumber];
		ref.isa = [_CORBAObjectRef class];
        _CORBAObjectRef *objRef =  (_CORBAObjectRef*)[_objects member:(id)&ref]; // May be released in other thread
#else
		_CORBAObjectRef *ref = [_CORBAObjectRef objectWithObjectID:[data objectID] address:address port:[data portNumber]];
        _CORBAObjectRef *objRef =  (_CORBAObjectRef*)[_objects member:ref]; // May be released in other thread
#endif
		if(!objRef && address && ([_addressSet containsObject:address] || [address isEqualToString:@"127.0.0.1"])){
#if USE_DEFS
			ref.address = nil;
			objRef = (_CORBAObjectRef*)[_objects member:&ref];
#else
			[ref->address release];
			ref->address = nil;
			objRef = (_CORBAObjectRef*)[_objects member:ref];
#endif
		}
		return objRef? objRef->object : nil;
    }else
        return nil;
}

+ (CORBAObject*)uniqueObjectForObject:(CORBAObject*)obj
{
    if(obj){
        CORBAObject* existingObj;
        [_hlock lock];
        existingObj = [self _lookupObject:obj];
        if(existingObj)
            obj = [[existingObj retain]autorelease];
        else{
			IDLProfileBody* data = [[obj profileWithTag:TAG_INTERNET_IOP]profileBody];
			_CORBAObjectRef *ref = [_CORBAObjectRef objectWithObjectID:[data objectID] 
				address:[data hostName] port:[data portNumber]];
			ref->object = obj;
			[_objects addObject:ref];
            [obj setIsUnique];
        }
        [_hlock unlock];
    }
    return obj;
}

+ _objectWithIORData:(NSData*)data
{
    IDLEncapsulation* stream = [IDLEncapsulation encapsulationWithData:(NSMutableData*)data];
    id obj = nil;
    // Check if ORB is set, otherwise set defaultORB
    if(![IDLDefinitionScope currentScope])
        [[ORB defaultORB]set];
    [stream decodeObject:&obj usingClass:self];
    return [self uniqueObjectForObject:obj];
}

+ objectWithIOR:(NSString*)iorString
{
    NSData* data = [ORB parseIORString:iorString];
    if(data)
        return [self _objectWithIORData:data];
    else
        return nil;
}

// Deprecated method
// UOL may have currently the following formats:
// file:<file-path> - path of IOR file
// name:<naming-service-path>
// http://<rest-of-URL> - URL of IOR file
// IOR:<rest-of-IOR> - IOR irself

+ objectWithUOL:(NSString*)uol
{
    if([uol hasPrefix:@"name:"]){
        id server = [ORB resolveInitialReferences:@"NameService"];
        NSString* namePath = [uol substringFromIndex:5];
        return [server resolve:[IDLNameComponent nameComponentsForPath:namePath]];
    }else
    if([uol hasPrefix:@"file:"]){
        NSString* filePath = [uol substringFromIndex:5];
        NSString* ior = [NSString stringWithContentsOfFile:filePath encoding:NSUTF8StringEncoding error:NULL];
        return ior?[self objectWithIOR:ior]:nil;
    }else
    if([uol hasPrefix:@"http:"]){
        NSURL* url = [NSURL URLWithString:uol];
        if(url){
            NSString* ior = [NSString stringWithContentsOfURL:url encoding:NSUTF8StringEncoding error:NULL];
            if(ior)
                return [self objectWithIOR:ior];
        }
        return nil;
    }else
    if([uol hasPrefix:@"IOR:"]){
        return [self objectWithIOR:uol];
    }else
        [NSException raise:InvalidFormatException format:@"invalid UOL format '%@'",uol];
    return nil;
}

+ objectWithURL:(NSString*)corbaUrl
{
// corbaloc:[iiop]:[[1.2@]<host>[:<port>]][/<object-key>]
// corbaname:[iiop]:[[1.2@]<host>[:<port>]][/<object-key>][#<string-name>]
// file://
// http://
// IOR:
    if([corbaUrl hasPrefix:@"IOR:"])
        return [self objectWithIOR:corbaUrl];
    else
    if([corbaUrl hasPrefix:@"http://"]){
        NSURL* url = [NSURL URLWithString:corbaUrl];
        if(url){
            NSString* newUrl = [NSString stringWithContentsOfURL:url encoding:NSUTF8StringEncoding error:NULL];
            if(newUrl)
                return [self objectWithURL:newUrl];
        }
        return nil;
    }else
    if([corbaUrl hasPrefix:@"file://"]){
        NSString* filePath = [corbaUrl substringFromIndex:7];
        NSString* newUrl = [NSString stringWithContentsOfFile:filePath encoding:NSUTF8StringEncoding error:NULL];
        return newUrl?[self objectWithURL:newUrl]:nil;
    }else
    if([corbaUrl hasPrefix:@"corbaloc:"]){
        IDLIOPProfile* profile = [IDLIOPProfile iopProfileWithCorbaloc:[corbaUrl substringFromIndex:9]];
        CORBAObject* object = [[[CORBAObject allocWithZone:[self zone]]initWithInterface:nil]autorelease];
   //	[profile addComponent:[IDLTaggedComponent codeSetsComponent]];
        [object addProfile:profile];
        return [CORBAObject uniqueObjectForObject:object];
    }else
    if([corbaUrl hasPrefix:@"corbaname:"]){
        IDLInterface* ifObj = (IDLInterface*)[[IDLDefinitionScope rootScope]
                definitionForNamePath:@"CosNaming/NamingContext"];
        NSRange r = {10, [corbaUrl length] - 10};
        NSRange r1 = [corbaUrl rangeOfString:@"#" options:0 range:r];
        NSString* locString;
        NSString* nameString = nil;
        IDLIOPProfile* profile;
        CORBAObject* naming = [[[CORBAObject allocWithZone:[self zone]]initWithInterface:ifObj]autorelease];
        if(r1.length > 0){
            r.length = r1.location - r.location;
            nameString = [corbaUrl substringFromIndex:r1.location+r1.length];
        }
        locString = [corbaUrl substringWithRange:r];
        profile = [IDLIOPProfile iopProfileWithCorbaloc:locString];
   //	[profile addComponent:[IDLTaggedComponent codeSetsComponent]];
        [naming addProfile:profile];
        naming = [CORBAObject uniqueObjectForObject:naming];
        if([nameString length] > 0)
            return [naming resolve:[IDLNameComponent nameComponentsForPath:nameString]];
        else
            return naming;
    }else
        [NSException raise:InvalidFormatException format:@"invalid URL format '%@'",corbaUrl];
    return nil;
}

- (NSString*) toString
{
    IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
    [stream encodeObject:self];
    return [ORB iorWithData:[stream data]];
}

extern NSString* VXVersionFromTypeId(NSString* typeId);

- (void) narrowInterface:(NSString*)typeId
{
    if(typeId){
        IDLInterface* obj = (IDLInterface*)[[IDLDefinitionScope currentScope] definitionForTypeId:typeId];
        if([obj isKindOfClass:[IDLInterface class]]){
            if(!VXVerifyInterfaceVersion || [VXVersionFromTypeId(typeId) isEqualToString:[obj version]])
                [self _setInterface:obj];
            else
                [NSException raise:IDLDefinitionException
                    format:@"Interface version '%@' does not match typeId '%@'",
                        [obj version], typeId];
        }else
        // We may just keep this object, like NameService. So we don't have and don't need the interface
            [self _setInterface:nil];
    }else
        [self _setInterface:nil];
}

- (void) narrowInterface
{
    [self narrowInterface:_typeId];
}

- (BOOL) implementsInterfaceNamed:(NSString*)name
{
    return [_interface implementsInterfaceNamed:name];
}

- (BOOL) isKindOfInterfaceNamed:(NSString*)name
{
    NSLog(@"Method '%@' is deprecated\n", NSStringFromSelector(_cmd));
    return [self implementsInterfaceNamed:name];
}

- (NSMutableDictionary*) invokeFunction:(IDLFunction*)fo with:(NSDictionary*)args
{
    NSMutableDictionary* dict = args? [NSMutableDictionary dictionaryWithDictionary:args] :
                            [NSMutableDictionary dictionary];
    VXInvocation* cinv = [self createInvocationFor:fo withArguments:dict];
    [cinv invoke];
    return dict;
}

- (NSMutableDictionary*) invokeMethod:(NSString*)methodName with:(NSDictionary*)args
{
    IDLFunction* fo = [[self interface]functionNamed:methodName];
    return [self invokeFunction:fo with:args];
}

- valueForAttribute:(NSString *)name
{
	IDLAttribute* attr = [[self interface]attributeNamed:name];
	if(attr){
		IDLFunction* fo = [attr getAttribute];
		NSDictionary* dict = [self invokeFunction:fo with:nil];
		return [dict objectForKey:@"$return"];
	}else
		[NSException raise:NSUndefinedKeyException format:@"Interface %@ has no attribute with name '%@'",
			[self interface], name];
	return nil;

}

- (void) setValue:value forAttribute:(NSString *)name
{
	IDLAttribute* attr = [[self interface]attributeNamed:name];
	if(attr){
		if(![attr isReadOnly]){
			IDLFunction* fo = [attr setAttribute];
			NSMutableDictionary* args = [NSMutableDictionary dictionary];
			[args setValue:value forKey:name];
			[self invokeFunction:fo with:args];
		}else
			[NSException raise:NSInvalidArgumentException 
				format:@"Cannot set attribute '%@'. The attribute is readonly", name];
	}else
		[NSException raise:NSUndefinedKeyException format:@"Interface %@ has no attribute with name '%@'",
			[self interface], name];
}


- (BOOL) respondsToSelector:(SEL)selector
{
    if(selector){
        if([super respondsToSelector:selector])
            return YES;
        else
        if([self isLocal])
            return [[self implementation]respondsToSelector:selector];
        else
            return ([[self interface]functionWithSelectorNamed:NSStringFromSelector(selector)] != nil);
    }else
        return NO;
}

- (BOOL) isLocal
{
    return (_implementation != nil);
}

// KeyValue coding

- (void) setType_id:(NSString*)typeId
{
    if(_typeId != typeId){
        [_typeId release];
        _typeId = [typeId retain];
        [self narrowInterface:_typeId];
    }
}

- (NSString*) type_id
{
    return _typeId;
}

- (void) setProfiles:(NSArray*)profiles
{
    if(((NSArray*)_profiles) != profiles){
        [_profiles release];
        _profiles = profiles? [[NSMutableArray arrayWithArray:profiles]retain] : nil;
    }
}
- (NSArray*)profiles
{
    return _profiles;
}

- (void) addProfile:profile
{
    if(!_profiles)
        _profiles = [[NSMutableArray alloc]init];
    [_profiles addObject:profile];
}

// Adds IOP profile to local object
- (void) addProfileWithKey:(NSData*)key
{
    IDLGIOPVersion* version = [[[IDLGIOPVersion alloc]init]autorelease];
    IDLProfileBody* profileData = [IDLProfileBody profileWith:version 
//                host:[(VXConnection*)[VXConnection connection]address]
				host:nil // for local object 'host' is always nil
                port:[[(VXConnection*)[VXConnection connection]port]unsignedShortValue]
                objectID:key];
    IDLIOPProfile* profile = [IDLIOPProfile iopProfileWithBody:profileData];
    [profile addComponent:[IDLTaggedComponent codeSetsComponent]];
    [self addProfile:profile];
	[[ORB notificationCenter]postNotificationName:ORBDidCreateObjectProfile object:self
		userInfo:[NSDictionary dictionaryWithObject:profile forKey:@"profile"]];
}

- (void) addProfileWithKey:(NSData*)key host:(NSString*)hostName port:(unsigned short)port 
        version:(IDLGIOPVersion*)version
{
    IDLProfileBody* profileData = [IDLProfileBody profileWith:
                version?version:[[[IDLGIOPVersion alloc]init]autorelease] 
                host:hostName
                port:port
                objectID:key];
    IDLIOPProfile* profile = [IDLIOPProfile iopProfileWithBody:profileData];
//    [profile addComponent:[IDLTaggedComponent codeSetsComponent]];
    [self addProfile:profile];
}

- (void) addProfileWithKey:(NSData*)key host:(NSString*)hostName port:(unsigned short)port
{
    IDLGIOPVersion* version = [[[IDLGIOPVersion alloc]init]autorelease];
    [self addProfileWithKey:key host:hostName port:port version:version];
}

static CORBAObject* lookupImplementation(id impl)
{
    if(impl){
        NSEnumerator* en = [_objects objectEnumerator];
        _CORBAObjectRef* ref = nil;
        while(ref = [en nextObject])
            if([ref->object implementation] == impl)
				break;
				
		return ref? ref->object : nil;
    }
    return nil;
}

static CORBAObject* lookupObjectID(NSData* objectKey)
{
    NSEnumerator *en = [_objects objectEnumerator];
    _CORBAObjectRef* ref = nil;
    while(ref = [en nextObject])
        if([ref->objectID isEqualToData:objectKey])
			break;
	return ref? ref->object : nil;
}

+ (CORBAObject*) lookupImplementation:impl
{
    id obj;
    [_hlock lock];
    obj = [[lookupImplementation(impl) retain]autorelease];
    [_hlock unlock];
    return obj;
}

+ (CORBAObject*) lookupObjectID:(NSData*)objectKey
{
    id obj;
    [_hlock lock];
    obj = [[lookupObjectID(objectKey) retain]autorelease];
    [_hlock unlock];
    return obj;
}

- (void) setImplementation:object
{
    [_implementation autorelease];
    _implementation = [object retain];
}

- (void) registerImplementation:object withObjectID:(NSData*)data
{
    if(_implementation == nil){
        NSException* exception = nil;
        [_hlock lock];
        @try{
			[self setImplementation:object];
			[self addProfileWithKey:data];
			[ORB registerObject:self];
			_CORBAObjectRef *ref = [_CORBAObjectRef objectWithObjectID:data 
				address:nil port:[[(VXConnection*)[VXConnection connection]port]unsignedShortValue]];
			ref->object = self;
			[_objects addObject:ref];
			[self setIsUnique];
        }@catch(NSException *localException){
			[self setImplementation:nil];
			[_profiles removeLastObject];
			exception = localException;
        }
        [_hlock unlock];
        if(exception)
            [exception raise];
    }else
        [NSException raise:CORBAAlreadyBoundException 
            format:@"object '%@' already has implementation registered", self];
}

- (void) registerImplementation:object withKey:(NSString*)stringKey
{
	NSData *data = [stringKey dataUsingEncoding:NSUTF8StringEncoding];
    [self registerImplementation:object withObjectID:data];
}

+ (CORBAObject*) objectForImplementation:impl withInterface:(IDLInterface*)ifObj
{
    CORBAObject* obj = nil;
    NSException* exception = nil;
    [_hlock lock];
    @try{
		obj = lookupImplementation(impl);
		if(!obj && impl){
			NSData* data = [CORBAObject objectIdForObject:impl];
			obj = [[[CORBAObject allocWithZone:[self zone]]initWithInterface:ifObj]autorelease];
			[obj setImplementation:impl];
			[obj addProfileWithKey:data];
			[obj setActivated:YES];
			_CORBAObjectRef *ref = [_CORBAObjectRef objectWithObjectID:data 
				address:nil port:[[(VXConnection*)[VXConnection connection]port]unsignedShortValue]];
			ref->object = obj;
			[_objects addObject:ref];
			[obj setIsUnique];
		}
    }@catch(NSException *localException){
		exception = localException;
    }
    [_hlock unlock];
    if(exception)
        [exception raise];
    return obj;
}

- (void) setActivated:(BOOL) flag
{
    _flags.activated = flag?YES:NO;
}

- (BOOL) isActivated
{
    return _flags.activated;
}

- (VXInvocation*) createInvocationFor:(IDLFunction*)fobj withArguments:(NSMutableDictionary*)args
{
    VXInvocation* inv = [[[VXInvocation allocWithZone:[ORB zone]]
        initWith:self function:fobj andArguments:args]autorelease];
    return inv;
}

- (NSError*) lastError
{
	[_hlock lock];
	NSError* error = [[_error retain]autorelease];
	[_hlock unlock];
	return error;
}

- (void) _setError:(NSError*)error
{
	[_hlock lock];
	NSError* oldError = _error;
	_error = [error retain];
	[oldError release];
	[_hlock unlock];
}

- (void) invalidateConnection:(NSNotification*)notification
{
    id conn = [notification object];
    if(conn == _connection){
        // Remove object from hash table so the new one can be created
        IDLProfileBody* data = [[self profileWithTag:TAG_INTERNET_IOP]profileBody];
#if USE_DEFS
		struct {@defs(_CORBAObjectRef)} ref;
		ref.objectID = [data objectID];
		ref.address = [data hostName];
		ref.port = [data portNumber];
		ref.isa = [_CORBAObjectRef class];
#else
		_CORBAObjectRef *ref = [_CORBAObjectRef objectWithObjectID:[data objectID] address:[data hostName] port:[data portNumber]];
#endif
        [_hlock lock];
#if USE_DEFS
		[_objects removeObject:(id)&ref];
#else
		[_objects removeObject:ref];
#endif
        [_hlock unlock];
		NSError* error = [[notification userInfo]objectForKey:@"error"];
		if(error)
			[self _setError:error];
        [ORB postNotification:[NSNotification notificationWithName:ORBInvalidateObject object:self userInfo:[notification userInfo]]];
    }
}

- (void) _connect
{
	if(!_connection){
		id profile = [self profileWithTag:TAG_INTERNET_IOP];
		if(profile){
			IDLProfileBody* data = [profile profileBody];
			NSString* host = [data hostName];
			unsigned short port = [data portNumber];
			NSError* error;
			_connection = [[VXConnection findConnectionWithListenPointHost:host port:port]retain];
			if(!_connection)
				_connection = [[VXConnection connectToHostNamed:host port:port timeout:_connectTimeout error:&error]retain];
			if(_connection){
				[self _setError:nil];
				[[NSNotificationCenter defaultCenter]addObserver:self
						selector:@selector(invalidateConnection:)
						name:VXInvalidateConnection
						object:nil];
//				[_connection setStringEncodingsFromComponent:[self taggedComponentWithTag:TAG_CODE_SETS]];
			}else
				// Do not post notification ORBInvalidateObject since the exception will be raised anyway
				// Just set the error
				[self _setError:error];
		}else{
			NSError* error = [NSError errorWithDomain:CORBAErrorDomain code:kENoProfile 
				userInfo:[NSDictionary dictionaryWithObject:@"No profile with tag TAG_INTERNET_IOP"
					forKey:NSLocalizedDescriptionKey]];
			[self _setError:error];
			NSLog(@"CORBAObject: No profile with tag TAG_INTERNET_IOP  in %@. Other tags are not supported. Invalidating object.\n", self);
		}
	}
	_flags.waitConnect = NO;
}

- connection
{
	if(!_connection && ![self isLocal]){
		if(!_flags.waitConnect){
			_flags.waitConnect = YES;
			[self _connect];
		}else
			[NSException raise:CORBACommunicationException format:@"connect in progress"];
	}
    return _connection;
}

- (void) _connectAndNotify
{
	[self _connect];
	if(_connection)
		[ORB postNotification:[NSNotification notificationWithName:CORBAObjectDidConnect object:self userInfo:nil]];
	else{
		NSDictionary* info = _error? [NSDictionary dictionaryWithObject:_error forKey:@"error"] : nil;
		[ORB postNotification:[NSNotification notificationWithName:CORBAObjectDidNotConnect object:self userInfo:info]];
	}
}

- (BOOL) connectInBackgroundAndNotify:(NSError**)error
{
	if(!_connection && ![self isLocal]){
		if(!_flags.waitConnect){
			_flags.waitConnect = YES;
			[NSThread detachNewThreadSelector:@selector(_connectAndNotify) toTarget:self withObject:nil];
			return YES;
		}
		if(error)
			*error = [NSError errorWithDomain:NSPOSIXErrorDomain code:EALREADY userInfo:nil];
	}else if(![self isLocal]){
		if(error)
			*error = [NSError errorWithDomain:NSPOSIXErrorDomain code:EISCONN userInfo:nil];
	}else if(error)
		*error = [NSError errorWithDomain:CORBAErrorDomain code:kELocalObject userInfo:nil];
	return NO;
}

- delegate
{
	return _delegate;
}

- (void) setDelegate:newDelegate
{
	if(newDelegate != _delegate){
		if(_delegate)
			[[NSNotificationCenter defaultCenter]removeObserver:_delegate name:nil object:self];
		_delegate = newDelegate;
		if(_delegate && [_delegate respondsToSelector:@selector(objectDidInvalidate:)])
			[[NSNotificationCenter defaultCenter]addObserver:_delegate
				selector:@selector(objectDidInvalidate:)
				name:ORBInvalidateObject
				object:self];
	}
}

- (BOOL) isConnected
{
    return (_connection != nil && [_connection isValid]);
}

- (NSData*)objectKey
{
    id profile = [self profileWithTag:TAG_INTERNET_IOP];
    if(profile){
        IDLProfileBody* data = [profile profileBody];
        return [data objectID];
    }else
        return nil;
}

- (NSData*)objectID
{
    return [self objectKey];
}

- (IDLGIOPVersion*) iiopVersion
{
    id profile = [self profileWithTag:TAG_INTERNET_IOP];
    if(profile){
        IDLProfileBody* data = [profile profileBody];
        return [data iiopVersion];
    }else
        return nil;
}
- profileWithTag:(int)tag
{
    NSEnumerator* en = [_profiles objectEnumerator];
    IDLIOPProfile* profile;
    while(profile = [en nextObject]){
        if([profile tag] == tag)
            return profile;
    }
    return nil;
}

- (NSNumber*) enumValueForName:(NSString*)name andKey:(NSString*)key 
{
    return [_interface enumValueForName:name andKey:key];
}

- (NSTimeInterval) replyTimeout
{
    return _replyTimeout;
}
- (void) setReplyTimeout:(NSTimeInterval)timeout
{
    _replyTimeout = timeout;
}
- (NSTimeInterval) connectTimeout
{
	return _connectTimeout;
}

- (void) setConnectTimeout:(NSTimeInterval)timeout
{
	_connectTimeout = timeout;
}

- (NSString*)description
{
	return [NSString stringWithFormat:@"<%@: 0x%x><%@>", NSStringFromClass([self class]), self, [self type_id]];
}

- (unsigned) hash
{
    IDLProfileBody* data = [[self profileWithTag:TAG_INTERNET_IOP]profileBody];
    NSString* host = [data hostName];
    unsigned port = [data portNumber];
    NSData* key = [data objectID];
    return [host hash] ^ port ^ [key hash] /*^ [_typeId hash]*/;
}

- (BOOL) isEqual:obj
{
    if(!obj || [self class] != [obj class])
        return NO;
    else{
        IDLProfileBody* data = [[self profileWithTag:TAG_INTERNET_IOP]profileBody];
        IDLProfileBody* objData = [[(CORBAObject*)obj profileWithTag:TAG_INTERNET_IOP]profileBody];
        NSString* host = [data hostName];
        NSString* objHost = [objData hostName];
        unsigned port = [data portNumber];
        unsigned objPort = [objData portNumber];
        NSData* key = [data objectID];
        NSData* objKey = [objData objectID];
        return (host && [objHost isEqualToString:host] || host == objHost) && 
            (key && [objKey isEqual:key] || key == objKey) &&
            port == objPort 
            /*&& [_typeId isEqualToString:obj->_typeId]*/ 
            /*&& _interface == obj->_interface*/;
    }
}

- (NSString*) host
{
    IDLProfileBody* data = [[self profileWithTag:TAG_INTERNET_IOP]profileBody];
    return [data hostName];
}

- (NSNumber*) port
{
    IDLProfileBody* data = [[self profileWithTag:TAG_INTERNET_IOP]profileBody];
    return [NSNumber numberWithUnsignedShort:[data portNumber]];
}

#pragma mark ---- Object Interface ----

- (BOOL) _isA:(NSString*)typeID
{
    IDLInterface* obj = (IDLInterface*)[[_interface rootScope]definitionForTypeId:typeID];
    return [_interface adoptsInterface:obj];
}

- (BOOL) _nonExistent // IDL: non_existent()
{
    return NO;
}

- (BOOL) isA:(NSString*)logicalTypeID // IDL: is_a(in type_id)
{
    if([self isLocal])
        return [self _isA:logicalTypeID];
    else{
        IDLFunction* fo = [[self interface]functionNamed:@"_is_a"];
        NSMutableDictionary* dict = [NSMutableDictionary dictionary];
        [dict setObject:logicalTypeID forKey:@"logical_type_id"];
        VXInvocation* cinv = [self createInvocationFor:fo withArguments:dict];
        [cinv invoke];
        NSNumber* rc = [dict objectForKey:@"$return"];
        return [rc boolValue];
    }
}

- (BOOL) nonExistent // IDL: non_existent()
{
    return NO;
}

+ (NSData*) objectIdForObject:object
{
    NSData* data = [NSData dataWithBytes:&object length:sizeof(id)];
	NSMutableData* uid = [NSMutableData dataWithData:_uuid];
	[uid appendData:data];
	return uid;
}

@end

@implementation NSObject(CORBAActivation)

- (void) activate
{
    NSData* data = [CORBAObject objectIdForObject:self];
    [self activateWithObjectID:data];
}

- (void) activateWithObjectID:(NSData*)key
{
    if([self respondsToSelector:@selector(interface)]){
        IDLInterface* ifObj = [(id<CORBAActivation>)self interface];
        if([ifObj isKindOfClass:[IDLInterface class]]){
            CORBAObject* obj = [[[CORBAObject allocWithZone:[self zone]]initWithInterface:ifObj]autorelease];
            [obj registerImplementation:self withObjectID:key];
            [obj setActivated:YES];
        }else
            [NSException raise:CORBANotActivatableException
                format:@"Object %@ cannot be activated", self];
            
    }else
        [NSException raise:CORBANotActivatableException
                format:@"Object %@ cannot be activated because it does not implement the method '-interface'", self];
}

- (void) activateWithObjectKey:(NSString*)key
{
	NSData* data = [key dataUsingEncoding:NSUTF8StringEncoding];
    [self activateWithObjectID:data];
}

- (void) deactivate
{
    CORBAObject* obj = [[[self objectReference]retain]autorelease];
    if(obj){
        [ORB unregisterObject:obj];
        [VXConnection removeObject:obj];
    }
}

- (CORBAObject*) objectReference
{
    return [CORBAObject lookupImplementation:self];
}

@end

#import <SystemConfiguration/SystemConfiguration.h>

@implementation CORBAObject(ORBIPMonitor)
#if !(TARGET_IPHONE_SIMULATOR || TARGET_OS_IPHONE)

static NSString* _gkey = nil;
static NSString* _key4 = nil;
static SCDynamicStoreRef _store = NULL;

static void updateAddresses(SCDynamicStoreRef store, NSArray* changedKeys, void* info);
static void _updateAddressAndNotify(BOOL flag);
static BOOL logEnabled = NO;

static void logSCError()
{
	int err = SCError();
	const char* desc =  SCErrorString(err);
	NSLog(@"*** System Configuration monitor failed, errno = %d, %s",
		err, desc? desc : "(unknown error)");
}

+ (void) setupIPMonitor
{
	NSString* str = [[NSUserDefaults standardUserDefaults]stringForKey:@"debug"];
	if(str && [str rangeOfString:@"M"].length > 0)
		logEnabled = YES;
	// State:/Network/Global/IPv4
	_gkey = (NSString*)SCDynamicStoreKeyCreateNetworkGlobalEntity(NULL, kSCDynamicStoreDomainState, kSCEntNetIPv4);
	// State:/Network/Service/[^/]+/IPv4
	_key4 = (NSString*)SCDynamicStoreKeyCreateNetworkServiceEntity(NULL, kSCDynamicStoreDomainState, kSCCompAnyRegex, kSCEntNetIPv4);
	_addressSet = [[NSMutableSet alloc]init];

    SCDynamicStoreContext   context = {0, NULL, NULL, NULL, NULL};
    context.info = self;
	_store = SCDynamicStoreCreate(NULL, (CFStringRef)@"ORBIPMonitor", (SCDynamicStoreCallBack)&updateAddresses, &context);
	if(!_store){
		logSCError();
		return;
	}
	
	if(SCDynamicStoreSetNotificationKeys(_store, 
			(CFArrayRef)[NSArray arrayWithObject:_gkey], 
			(CFArrayRef)[NSArray arrayWithObject:_key4]))
	{
		CFRunLoopSourceRef rls = SCDynamicStoreCreateRunLoopSource(NULL, _store, 0);
		NSAssert1(rls != nil, @"ORBIPMonitor cannot create run loop source, errno = %d", SCError());
		CFRunLoopAddSource([[NSRunLoop currentRunLoop]getCFRunLoop], rls, kCFRunLoopCommonModes);
	}else{
		logSCError();
		return;
	}
	_updateAddressAndNotify(NO);
}

static void updateAddresses(SCDynamicStoreRef store, NSArray* changedKeys, void* info)
{
	if(logEnabled)
		NSLog(@"IPMonitor keys changed: %@", changedKeys);
	_updateAddressAndNotify(YES);
}

static void _updateAddressAndNotify(BOOL flag)
{
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
	NSDictionary* global = (NSDictionary*)SCDynamicStoreCopyValue(_store, (CFStringRef)_gkey);
	NSString* serviceId = global? [global objectForKey:(NSString*)kSCDynamicStorePropNetPrimaryService] : nil;
	NSString* address = nil;
	NSDictionary* values = (NSDictionary*)SCDynamicStoreCopyMultiple(_store, NULL, (CFArrayRef)[NSArray arrayWithObject:_key4]);
	NSMutableSet* set = [NSMutableSet set];
	NSEnumerator* en = [values keyEnumerator];
	NSString* key;
	while(key = [en nextObject]){
		NSArray* array = [[values objectForKey:key]objectForKey:(NSString*)kSCPropNetIPv4Addresses];
		if(array)
			[set addObjectsFromArray:array];
	}
	[values release];
	if(serviceId){
		NSString* entityKey4 = (NSString*)SCDynamicStoreKeyCreateNetworkServiceEntity(NULL, kSCDynamicStoreDomainState, (CFStringRef)serviceId, kSCEntNetIPv4);
		values = (NSDictionary*)SCDynamicStoreCopyValue(_store, (CFStringRef)entityKey4);
		NSArray* addresses = [values objectForKey:(NSString*)kSCPropNetIPv4Addresses];
		address = [addresses count] > 0? [[addresses objectAtIndex:0]copy] : nil;
		[values release];
		[entityKey4 release];
	}
	
	BOOL addrChanged = address != nil && ![_address isEqualToString:address] || address == nil && _address != nil;
	NSString* oldAddress = _address;

	[_hlock lock];
	[_addressSet removeAllObjects];
	[_addressSet unionSet:set];
	if(addrChanged)
		_address = address;
	[_hlock unlock];
	
	if(logEnabled)
		NSLog(@"IPMonitor: address set to: %@ %@", _address, _addressSet);
	if(addrChanged){
		if(flag){
			NSMutableDictionary* info = [NSMutableDictionary dictionary];
			if(_address)
				[info setObject:_address forKey:@"address"];
			if(oldAddress)
				[info setObject:oldAddress forKey:@"oldAddress"];
			[[NSNotificationCenter defaultCenter]postNotificationName:@"ORBLocalAddressDidChange"
				object:[ORB class] userInfo:info];
		}
		[oldAddress release];
	}else
		[address release];
	
	[global release];
	[pool release];
}
#endif

@end

@implementation _CORBAObjectRef
- initWithObjectID:(NSData*)objId address:(NSString*)addr port:(unsigned)portNum
{
	self = [super init];
	objectID = [objId retain];
	address = [addr retain];
	port = portNum;
	return self;
}

- (void) dealloc
{
	[objectID release];
	[address release];
	[super dealloc];
}

+ objectWithObjectID:(NSData*)objId address:(NSString*)addr port:(unsigned)port
{
	return [[[self alloc]initWithObjectID:objId address:addr port:port]autorelease];
}

- (unsigned) hash
{
	return [objectID hash] ^ [address hash] ^ port;
}

- (BOOL) isEqual:(_CORBAObjectRef*)obj
{
	return [obj isMemberOfClass:[_CORBAObjectRef class]] &&
		(obj->objectID && [objectID isEqual:obj->objectID] || objectID == obj->objectID) &&
		(obj->address && [address isEqualToString:obj->address] || address == obj->address) &&
		port == obj->port;
}

@end

