/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "VXMessage.h"
#import "VXMessageHeader.h"
#import "IDLOctetStream.h"
#import "VXConnection.h"
#import "CORBAObject.h"
#import "IDLDefinitionScope.h"
#import "IDLDefinition.h"

@implementation VXMessage

static int _debugSend = 0;
static int _debugRecv = 0;

+ (void) initialize
{
    if([VXMessage class] == self){
        NSString* str = [[NSUserDefaults standardUserDefaults]stringForKey:@"debug"];
        if(str && [str rangeOfString:@"s"].length > 0)
            _debugSend = YES;
        if(str && [str rangeOfString:@"r"].length > 0)
            _debugRecv = YES;
    }
}

- initWithMessageType:(unsigned)msgType iiopVersion:(IDLGIOPVersion*)version
{
    self = [super init];
    _stream = [[IDLOctetStream alloc]initWithOffset:0];
    _header = [[VXMessageHeader alloc]initWithMessageType:msgType iiopVersion:version stream:_stream];
    [_stream setIIOPVersion:version];
	[_stream setMessage:self];
    return self;
}

- initWithStream:(IDLOctetStream*)stream header:(VXMessageHeader*)header
{
	self = [super init];
	_stream = [stream retain];
	_header = [header retain];
	[_stream setCursorPosition:MSG_HEADER_SIZE];
	return self;
}

- (void) dealloc
{
    [_header release];
    [_stream release];
    [super dealloc];
}

- (IDLOctetStream*) stream
{
    return _stream;
}

- (VXMessageHeader*) messageHeader
{
    return _header;
}

- (unsigned long) minorVersion
{
    return [_header minorVersion];
}

- (unsigned long) majorVersion
{
    return [_header majorVersion];
}

- (IDLGIOPVersion*)iiopVersion
{
    return [_header iiopVersion];
}

- (unsigned long) messageType
{
    return [_header messageType];
}

+ requestWithIIOPVersion:(IDLGIOPVersion*)iiopVersion
{
    return [[[self alloc]initWithMessageType:GIOP_RequestType iiopVersion:iiopVersion]autorelease];
}

- (void) sendWithConnection:connection
{
	[_header set];
	[(VXConnection*)connection writeData:[_stream data]];
    if(_debugSend){
        NSLog(@"Sent message:\n%@\n%@\n\n",
            [_header description],
            [[_stream data]description]);
    }
}

- _peepHeader
{
    unsigned major = [_header majorVersion];
    unsigned minor = [_header minorVersion];
    NSString* path;
    IDLDefinition* def = nil;
    switch([_header messageType]){
        case GIOP_RequestType:
            path = [NSString stringWithFormat:@"GIOP/RequestHeader_%d_%d",major,minor];
            break;
        case GIOP_ReplyType:
            path = [NSString stringWithFormat:@"GIOP/ReplyHeader_%d_%d",major,minor];
            break;
        case GIOP_CancelRequestType:
            path = @"GIOP/CancelRequestHeader";
            break;
        case GIOP_LocateRequestType:
            path = [NSString stringWithFormat:@"GIOP/LocateRequestHeader_%d_%d",major,minor];
            break;
        case GIOP_LocateReplyType:
            path = [NSString stringWithFormat:@"GIOP/LocateReplyHeader_%d_%d",major,minor];
            break;
        case GIOP_CloseConnectionType:
        case GIOP_MessageErrorType:
        default:
            return nil;
    }
    def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
    if(def){
        id obj = nil;
        [_stream peepStruct:&obj usingObjectCoder:def];
        return obj;
    }else
        return nil;
}

- peepHeader
{
    if([_stream isInputStream])
        return [self _peepHeader];
    else
        return nil;
}

- (void) alignStreamBody
{
    if([self minorVersion] >= 2)
        [_stream setNeedsAlign:YES];
}

- (long) requestId
{
    id header = [self peepHeader];
    if(header)
        return [[header valueForKey:@"request_id"]intValue];
    else
        return 0;
}

- invocationObject
{
	return _invocation;
}

- (void) setInvocationObject:object
{
	_invocation = object; // invocation object owns the message
	if(_invocation){
		[_stream setStringEncoding:[_invocation stringEncoding]];
		[_stream setWStringEncoding:[_invocation wstringEncoding]];
	}
}

@end
